/// <reference types="w3c-web-serial" />
import { ISerialTransport, ISerialOptions } from "./ISerialTransport.js";
import { ITrace } from "../utils/ITrace.js";
/**
 * Options for device serialPort.
 * @interface SerialOptions
 */
export interface SerialOptions extends ISerialOptions {
    /**
     * The number of data bits per frame. Either 7 or 8.
     * @type {number | undefined}
     */
    dataBits?: number | undefined;
    /**
     * The number of stop bits at the end of a frame. Either 1 or 2.
     * @type {number | undefined}
     */
    stopBits?: number | undefined;
    /**
     * The parity mode: none, even or odd
     * @type {ParityType | undefined}
     */
    parity?: ParityType | undefined;
    /**
     * A positive, non-zero value indicating the size of the read and write buffers that should be created.
     * @type {number | undefined}
     */
    bufferSize?: number | undefined;
    /**
     * The flow control mode: none or hardware.
     * @type {FlowControlType | undefined}
     */
    flowControl?: FlowControlType | undefined;
}
/**
 * Wrapper class around Webserial API to communicate with the serial device.
 * @param {typeof import("w3c-web-serial").SerialPort} device - Requested device prompted by the browser.
 * @param {boolean} tracing - Enable communication tracing
 *
 * ```
 * const port = await navigator.serial.requestPort();
 * ```
 */
export declare class WebSerialTransport implements ISerialTransport {
    device: SerialPort;
    private tracer?;
    leftOver: Uint8Array;
    private reader;
    constructor(device: SerialPort, tracer?: ITrace | undefined);
    /**
     * Request the serial device vendor ID and Product ID as string.
     * @returns {string} Return the device VendorID and ProductID from SerialPortInfo as formatted string.
     */
    getInfo(): string;
    /**
     * Request the serial device product id from SerialPortInfo.
     * @returns {number | undefined} Return the product ID.
     */
    getPID(): number | undefined;
    /**
     * Write binary data to device using the WebSerial device writable stream.
     * @param {Uint8Array} outData 8 bit unsigned data array to write to device.
     */
    write(outData: Uint8Array): Promise<void>;
    /**
     * Read from serial device without slip formatting.
     * @param {number} timeout Read timeout in milliseconds (ms)
     * @param {number} minData Minimum packet array length
     * @returns {Promise<Uint8Array>} 8 bit unsigned data array read from device.
     */
    read(timeout?: number, minData?: number): Promise<Uint8Array>;
    _DTR_state: boolean;
    /**
     * Send the RequestToSend (RTS) signal to given state
     * # True for EN=LOW, chip in reset and False EN=HIGH, chip out of reset
     * @param {boolean} state Boolean state to set the signal
     */
    setRTS(state: boolean): Promise<void>;
    /**
     * Send the dataTerminalReady (DTS) signal to given state
     * # True for IO0=LOW, chip in reset and False IO0=HIGH
     * @param {boolean} state Boolean state to set the signal
     */
    setDTR(state: boolean): Promise<void>;
    /**
     * Connect to serial device using the Webserial open method.
     * @param {SerialOptions} serialOptions Serial Options for ESP Loader class.
     */
    connect(serialOptions: SerialOptions): Promise<void>;
    sleep(ms: number): Promise<unknown>;
    /**
     * Wait for a given timeout ms for serial device unlock.
     * @param {number} timeout Timeout time in milliseconds (ms) to sleep
     */
    waitForUnlock(timeout: number): Promise<void>;
    /**
     * Disconnect from serial device by running SerialPort.close() after streams unlock.
     */
    disconnect(): Promise<void>;
}
