/* global SerialPort, ParityType, FlowControlType */
import { hexConvert } from "../utils/hex.js";
import { appendArray } from "../utils/convert.js";
/**
 * Wrapper class around Webserial API to communicate with the serial device.
 * @param {typeof import("w3c-web-serial").SerialPort} device - Requested device prompted by the browser.
 * @param {boolean} tracing - Enable communication tracing
 *
 * ```
 * const port = await navigator.serial.requestPort();
 * ```
 */
export class WebSerialTransport {
    constructor(device, tracer) {
        this.device = device;
        this.tracer = tracer;
        this.leftOver = new Uint8Array(0);
        this._DTR_state = false;
    }
    /**
     * Request the serial device vendor ID and Product ID as string.
     * @returns {string} Return the device VendorID and ProductID from SerialPortInfo as formatted string.
     */
    getInfo() {
        const info = this.device.getInfo();
        return info.usbVendorId && info.usbProductId
            ? `WebSerial VendorID 0x${info.usbVendorId.toString(16)} ProductID 0x${info.usbProductId.toString(16)}`
            : "";
    }
    /**
     * Request the serial device product id from SerialPortInfo.
     * @returns {number | undefined} Return the product ID.
     */
    getPID() {
        return this.device.getInfo().usbProductId;
    }
    /**
     * Write binary data to device using the WebSerial device writable stream.
     * @param {Uint8Array} outData 8 bit unsigned data array to write to device.
     */
    async write(outData) {
        if (this.device.writable) {
            const writer = this.device.writable.getWriter();
            if (this.tracer) {
                this.tracer.trace("Write bytes");
                this.tracer.trace(`Write ${outData.length} bytes: ${hexConvert(outData)}`);
            }
            await writer.write(outData);
            writer.releaseLock();
        }
    }
    /**
     * Read from serial device without slip formatting.
     * @param {number} timeout Read timeout in milliseconds (ms)
     * @param {number} minData Minimum packet array length
     * @returns {Promise<Uint8Array>} 8 bit unsigned data array read from device.
     */
    async read(timeout = 0, minData = 0) {
        if (this.leftOver.length != 0) {
            const p = this.leftOver;
            this.leftOver = new Uint8Array(0);
            return p;
        }
        if (!this.device.readable) {
            return this.leftOver;
        }
        this.reader = this.device.readable.getReader();
        let t;
        let packet = this.leftOver;
        try {
            if (timeout > 0) {
                t = setTimeout(() => {
                    if (this.reader) {
                        this.reader.cancel();
                    }
                }, timeout);
            }
            do {
                const { value, done } = await this.reader.read();
                if (done) {
                    this.leftOver = packet;
                    throw new Error("Timeout");
                }
                if (this.tracer) {
                    this.tracer.trace("Raw Read bytes");
                    this.tracer.trace(`Read ${value.length} bytes: ${hexConvert(value)}`);
                }
                packet = appendArray(packet, value);
            } while (packet.length < minData);
            return packet;
        }
        finally {
            if (timeout > 0) {
                clearTimeout(t);
            }
            this.reader.releaseLock();
        }
    }
    /**
     * Send the RequestToSend (RTS) signal to given state
     * # True for EN=LOW, chip in reset and False EN=HIGH, chip out of reset
     * @param {boolean} state Boolean state to set the signal
     */
    async setRTS(state) {
        await this.device.setSignals({ requestToSend: state });
        // # Work-around for adapters on Windows using the usbser.sys driver:
        // # generate a dummy change to DTR so that the set-control-line-state
        // # request is sent with the updated RTS state and the same DTR state
        // Referenced to esptool.py
        await this.setDTR(this._DTR_state);
    }
    /**
     * Send the dataTerminalReady (DTS) signal to given state
     * # True for IO0=LOW, chip in reset and False IO0=HIGH
     * @param {boolean} state Boolean state to set the signal
     */
    async setDTR(state) {
        this._DTR_state = state;
        await this.device.setSignals({ dataTerminalReady: state });
    }
    /**
     * Connect to serial device using the Webserial open method.
     * @param {SerialOptions} serialOptions Serial Options for ESP Loader class.
     */
    async connect(serialOptions) {
        await this.device.open({
            baudRate: serialOptions.baudRate || 115200,
            dataBits: serialOptions === null || serialOptions === void 0 ? void 0 : serialOptions.dataBits,
            stopBits: serialOptions === null || serialOptions === void 0 ? void 0 : serialOptions.stopBits,
            bufferSize: serialOptions === null || serialOptions === void 0 ? void 0 : serialOptions.bufferSize,
            parity: serialOptions === null || serialOptions === void 0 ? void 0 : serialOptions.parity,
            flowControl: serialOptions === null || serialOptions === void 0 ? void 0 : serialOptions.flowControl,
        });
        this.leftOver = new Uint8Array(0);
    }
    async sleep(ms) {
        return new Promise((resolve) => setTimeout(resolve, ms));
    }
    /**
     * Wait for a given timeout ms for serial device unlock.
     * @param {number} timeout Timeout time in milliseconds (ms) to sleep
     */
    async waitForUnlock(timeout) {
        while ((this.device.readable && this.device.readable.locked) ||
            (this.device.writable && this.device.writable.locked)) {
            await this.sleep(timeout);
        }
    }
    /**
     * Disconnect from serial device by running SerialPort.close() after streams unlock.
     */
    async disconnect() {
        var _a, _b;
        if ((_a = this.device.readable) === null || _a === void 0 ? void 0 : _a.locked) {
            await ((_b = this.reader) === null || _b === void 0 ? void 0 : _b.cancel());
        }
        await this.waitForUnlock(400);
        this.reader = undefined;
        await this.device.close();
    }
}
