import { ISerialTransport } from "../transport/ISerialTransport.js";
import { ITrace } from "./ITrace.js";
/**
 * Slip reader output packet and left over as Uint8Array.
 * @interface SlipReaderOutput
 */
export interface SlipReaderOutput {
    /**
     * Formatted packet using SLIP escape sequences.
     * @type {Uint8Array}
     */
    packet: Uint8Array;
    /**
     * Updated left over bytes from read operation
     * @type {Uint8Array}
     */
    newLeftOver: Uint8Array;
}
/**
 * Class to handle SLIP read and write serial methods.
 * @param {ISerialTransport} transport Transport object with raw read and write serial methods
 * @param {boolean} enableSlipRead Enable or disable read SLIP data formatting.
 * @param {ITrace} trace Object that log or trace all serial messages.
 */
export declare class Slip {
    private transport;
    enableSlipRead: boolean;
    private tracer?;
    constructor(transport: ISerialTransport, enableSlipRead?: boolean, tracer?: ITrace | undefined);
    /**
     * Format data packet using the Serial Line Internet Protocol (SLIP).
     * @param {Uint8Array} data Binary unsigned 8 bit array data to format.
     * @returns {Uint8Array} Formatted unsigned 8 bit data array.
     */
    encode(data: Uint8Array): Uint8Array;
    /**
     * Take a data array and return the first well formed packet after
     * replacing the escape sequence. Reads at least 8 bytes.
     * @param {Uint8Array} data Unsigned 8 bit array from the device read stream.
     * @returns {SlipReaderOutput} packet Formatted packet using SLIP escape sequences.
     */
    decode(data: Uint8Array): SlipReaderOutput;
    /**
     * Read from serial device using the device ReadableStream.
     * @param {number} timeout Read timeout number
     * @param {number} minData Minimum packet array length
     * @returns {Promise<Uint8Array>} 8 bit unsigned data array read from device.
     */
    read(timeout?: number, minData?: number): Promise<Uint8Array>;
}
